#include <stdio.h>
#include <string.h>
#include <alloc.h>
#include <mem.h>

#include "unSNAP.hpp"

const word loaderSize = 53;

const word PAGE_SIZE = 0x4000;
byte page[PAGE_SIZE];

TSnapState ss;
SnapFormat format;
SnapType type;

byte pageTable[8] = {0, 0, 0, 0, 0, 0, 0, 0};

FILE *snap;
char name[9];
long fileSize;

word checkSum(byte *ptr)
{
 word sum = 0;
 for(int i = 0; i < 15; i++) sum += ptr[i];
 return 257*sum + 105;
}

int getSnapState(void)
{
 if(format == Z80)
 {
  TZ80Header rs;
  fread(&rs, sizeof(TZ80Header), 1, snap);
  
  if(rs.pc)
  {
   ss.hmode = mode48;    // ଠ z80 v1.xx
   type = Z80_1xx;
   ss.isPacked = (rs.flags1&0x20)>>5;
   ss.pc = rs.pc;
  }
  else
  {
   word addHeaderSize;  // ଠ z80 v2.xx  v3.xx
   fread(&addHeaderSize, sizeof(word), 1, snap);
   
   switch(addHeaderSize)
   {
    case 23:
              type = Z80_2xx;
              break;
    case 54:
              type = Z80_3xx;
              break;
    default:
              puts("Invalid Z80 header");
              return 0;
   }
   
   byte addHeader[54];
   fread(addHeader, addHeaderSize, 1, snap);
   ss.pc = *(word*)addHeader;
   ss.port7ffd = addHeader[3];
   byte hMode = addHeader[2];
   
   ss.hmode = mode128;
   if(hMode < 3 ) ss.hmode = mode48;
   if((type == Z80_2xx && hMode == 2)||(type == Z80_3xx && hMode == 3))
   {
    puts("SamRam hardware mode not supported");
    return 0;
   }    
  }
  ss.hl = rs.hl; ss.de = rs.de; ss.bc = rs.bc;
  ss._hl = rs._hl; ss._de = rs._de; ss._bc = rs._bc;
  ss.ix = rs.ix; ss.iy = rs.iy; ss.sp = rs.sp;
  ss.a = rs.a; ss.f = rs.f; ss._a = rs._a; ss._f = rs._f;
  ss.i = rs.i;
  
  ss.istatus = rs.istatus;
  ss.imode = rs.flags2 & 3;
  ss.bcolor = ( rs.flags1 >> 1 ) & 7;
  ss.bit7r = ( rs.flags1 & 1 ) << 7;
  ss.r = rs.r & 0x7f; ss.r |= ss.bit7r;
 }
 else
 {
  ss.hmode = mode128;
  switch(fileSize)
  {
   case 49179L:
             ss.hmode = mode48;
             type = SNA48;
             break;
   case 131103L:
             type = SNA128_1;
             break;
   case 147487L:
             type = SNA128_2;
             break;
   default:
             return 0;
  }
  TSNAHeader rs;
  fread(&rs, sizeof(TSNAHeader), 1, snap);
  ss.pc = 0;
  if(type != SNA48)
  {
   fseek(snap, 49179L, SEEK_SET);
   fread(&ss.pc, 2, 1, snap);
   fread(&ss.port7ffd, 1, 1, snap);
  }
  ss.hl = rs.hl; ss.de = rs.de; ss.bc = rs.bc;
  ss._hl = rs._hl; ss._de = rs._de; ss._bc = rs._bc;
  ss.ix = rs.ix; ss.iy = rs.iy; ss.sp = rs.sp;
  ss.a = rs.a; ss.f = rs.f; ss._a = rs._a; ss._f = rs._f;
  ss.i = rs.i; ss.r = rs.r; ss.bit7r = rs.r&0x80;
  ss.imode = rs.imode; ss.istatus = (rs.istatus&0x04)>>2;
  ss.bcolor = rs.bcolor;
  fseek(snap, sizeof(TSNAHeader), SEEK_SET);
 }
 return 1;
}

void writeRGSFile(void)
{
 char RGSFileName[13];
 sprintf(RGSFileName, "%s.RGS", strupr(name));
 
 FILE *RGSFile = fopen(RGSFileName, "wt");
 if(!RGSFile)
 {
  printf("Can't create RGS file %s", RGSFileName);
  return;
 }
 fprintf(RGSFile, "Snapshot type is ");
 printf("Snapshot type is ");
 if(format == SNA)
 {
  fprintf(RGSFile, "SNA\n");
  printf("SNA\n");
 }
 else
 {
  fprintf(RGSFile, "Z80 v");
  printf("Z80 v");
  switch(type)
  {
   case Z80_1xx:
             fprintf(RGSFile, "1.xx (");
             printf("1.xx (");
             if(ss.isPacked)
             {
              fprintf(RGSFile, "packed)\n");
              printf("packed)\n");
             }
             else
             {
              fprintf(RGSFile, "not packed)\n");
              printf("not packed)\n");
             }
             break;
   case Z80_2xx:
             fprintf(RGSFile, "2.xx\n");
             printf("2.xx\n");
             break;
   case Z80_3xx:
             fprintf(RGSFile, "3.xx\n");
             printf("3.xx\n");
             break;
  }
 }
 fprintf(RGSFile, "Hardware mode is %s\n\n", ss.hmode == mode48 ? "48k" : "128k");
 printf("Hardware mode is %s\n\n", ss.hmode == mode48 ? "48k" : "128k");
 
 fprintf(RGSFile, "HL: #%04X\tHL': #%04X\n", ss.hl, ss._hl);
 fprintf(RGSFile, "DE: #%04X\tDE': #%04X\n", ss.de, ss._de);
 fprintf(RGSFile, "BC: #%04X\tBC': #%04X\n", ss.bc, ss._bc);
 fprintf(RGSFile, "IX: #%04X\tIY : #%04X\n", ss.ix, ss.iy);
 fprintf(RGSFile, "SP: #%04X", ss.sp);
 
 if(type != SNA48) fprintf(RGSFile, "\tPC : #%04X", ss.pc);
 fprintf(RGSFile, "\n");
 
 fprintf(RGSFile, "A : #%02X\t\tA' : #%02X\n", ss.a, ss._a);
 fprintf(RGSFile, "F : #%02X\t\tF' : #%02X\n", ss.f, ss._f);
 fprintf(RGSFile, "R : #%02X\t\tI  : #%02X\n\n", ss.r, ss.i);
 fprintf(RGSFile, "Border color: %02X\n", ss.bcolor);
 fprintf(RGSFile, "Interrupts %s\n", ss.istatus ? "enabled" : "disabled");
 fprintf(RGSFile, "Interrupt mode is IM%d\n", ss.imode);
 if(ss.hmode == mode128) fprintf(RGSFile, "Last out to #7ffd is #%02X\n", ss.port7ffd);
 fclose(RGSFile);
}

void savePage(int pageNum)
{
 static int index = 0;
 if(pageNum < 0 || pageNum > 7)
 {
  puts("Invalid page found");
  return;
 }
 HoHdr pageHoHdr = {{"        ", 'C', 0xc000, 0x4000 }, ' ', 0x40};
 char pageFileName[12];
 printf("Page %02x found\n", pageNum);
 strncpy(pageHoHdr.file.name, name, strlen(name));
 pageHoHdr.file.type = '0' + pageNum;
 pageHoHdr.CheckSum = checkSum((byte*)&pageHoHdr);
 
 if(pageNum == 5)
 {
  sprintf(pageFileName, "%s.$7", name);
 }
 else
 {
  int isEmpty = 1;
  for(int i = 0; i < 0x4000; i++) if(page[i]) { isEmpty = 0; break; }
  if(isEmpty)
  {
   puts("This page is empty. Skiped...");
   return;
  }
  sprintf(pageFileName, "%s.$%c", name, (char)('0' + index));
  pageTable[index++] = pageNum | 0x10;
 }
 
 FILE *pageFile = fopen(pageFileName, "wb");
 if(!pageFile)
 {
  printf("Can't create file %s\n", pageFileName);
 }
 else
 {
  fwrite(&pageHoHdr, sizeof(HoHdr), 1, pageFile);
  fwrite(page, PAGE_SIZE, 1, pageFile);
  fclose(pageFile);
  printf("File %s created\n", pageFileName);
 }
}

void readSavePage(int pageNum)
{
 fread(page, PAGE_SIZE, 1, snap);
 savePage(pageNum);
}

void saveSNA48pages(void)
{
 readSavePage(5);
 readSavePage(2);
 readSavePage(0);
}

void saveSNA128pages(void)
{
 readSavePage(5);
 readSavePage(2);
 fread(page, PAGE_SIZE, 1, snap);
 
 fseek(snap, 4l, SEEK_CUR); // ய⨫ PC  port7ffd
 
 int pageNum[6] = {0, 1, 3, 4, 6, 7};
 
 if(type == SNA128_1)
 {
  int pagedPage = ss.port7ffd & 0x03;
  for(int i = 5; i; i--) if(pageNum[i] <= pagedPage) pageNum[i] = pageNum[i-1];
  pageNum[0] = pagedPage;
 }
 else
 {
  fread(page, PAGE_SIZE, 1, snap);
 }
 savePage(pageNum[0]);
 for(int i = 1; i < 6; i++) readSavePage(pageNum[i]);
}

void putByte(byte b)
{
 static byte *to = page;
 static int curPage = 0;
 int pageNum[] = {5, 2, 0};
 
 *to++ = b;
 if(to-page < PAGE_SIZE) return;
 // ࠭ ⠪ ஢ઠ, ୮   㤥 모
 if(curPage == 3)
 {
  puts("Invalid snapshot. Too many pages...");
 }
 to = page;
 savePage(pageNum[curPage]);
 curPage++;
}

void saveZ80_1pages(void)
{
 if(ss.isPacked)
 {
  //  snapshot Z80 v1.xx
  byte *fileBuf = (byte*) malloc(fileSize-sizeof(TZ80Header));
  if(!fileBuf)
  {
   fclose(snap);
   puts("Not enough memory");
   return;
  }
  else
  {
   fread(fileBuf, fileSize-sizeof(TZ80Header), 1, snap);
   byte *from = fileBuf;
   while(1)
   {
    if(from[0] == 0 && from[1] == 0xED && from[2] == 0xED && from[3] == 0) break;
    if(from[0] == 0xED && from[1] == 0xED)
    {
     byte noRepetitions = from[2];
     byte pattern = from[3];
     from += 4;
     for(byte i = 0; i < noRepetitions; i++) putByte(pattern);
    }
    else
    {
     putByte(*from++);
    }
   }
  }
 }
 else
 {
  //  snapshot Z80 v1.xx
  saveSNA48pages();
 }
}

void decodePage(byte *from, byte *to, int packedSize)
{
 for(int i = 0; i < packedSize; i++)
 {
  if(from[0] == 0xED && from[1] == 0xED)
  {
   byte noRepetitions = from[2];
   byte pattern = from[3];
   from += 4;
   for(byte j = 0; j < noRepetitions; j++) *to++ = pattern;
   i += 3;
  }
  else
  {
   *to++ = *from++;
  }
 }
 if(to-page != 0x4000) puts("Unpacking error...");
}

int translate(byte pageNum)
{
 int pages48[12] = {-1, -1, -1, -1, 2, 0, -1, -1, 5, -1, -1, -1};
 int pages128[12] = {-1, -1, -1,  0, 1, 2,  3,  4, 5,  6,  7, -1};
 if(pageNum > 11) return -1;
 if(ss.hmode == mode48)
  return pages48[pageNum];
 else
  return pages128[pageNum];
}

void saveZ80_23pages(void)
{
 int noPages = (ss.hmode == mode48) ? 3 : 8;
 byte packedData[2*PAGE_SIZE];
 for(int i = 0; i < noPages; i++)
 {
  word packedSize;
  fread(&packedSize, sizeof(word), 1, snap);
  if(feof(snap)) return;
  byte virtualPageNum;
  fread(&virtualPageNum, sizeof(byte), 1, snap);
  fread(packedData, packedSize, 1, snap);
  decodePage(packedData, page, packedSize);
  savePage(translate(virtualPageNum));
 }
}

void main(int argc, char **argv)
{
 puts("\nSnapshot to Hobeta Converter v0.2b\n(c) 1998-9 by HalfElf, N.Novgorod, RU\n");
 if(argc < 2 || argc > 3)
 {
  puts("Usage: unSNAP [-xxxx] <snapshot>\n"
       "  xxxx - address to place loader\n");
  return;
 }

 word loaderAddr = 0x4100;
 char *fullName = argv[1];
 if(argc == 3 && (argv[1][0] == '-' || argv[1][0] == '/'))
 {
  sscanf(argv[1]+1, "%04x", &loaderAddr);
  if(loaderAddr < 0x4000 || loaderAddr > 0xFFFF - loaderSize || (loaderAddr > 0x7F00 && loaderAddr < 0x8000))
  {
   loaderAddr = 0x4100;
   puts( "Invalid switch" );
  }
  fullName = argv[2];
 }
 // ஢ਬ ७
 char *ptrToExt = fullName + strlen(fullName) - 4;
 if(!stricmp(ptrToExt, ".z80"))
 {
  format = Z80;
 }
 else
 {
  if(!stricmp(ptrToExt, ".sna"))
  {
   format = SNA;
  }
  else
  {
   puts(".SNA or .Z80 required");
   return;
  }
 }
 
 // 뤥塞  䠩
 char *tmpPtr = fullName;
 char *ptrToName = fullName;
 while(*tmpPtr)
 {
  if(*tmpPtr == '\\' || *tmpPtr==':') ptrToName = tmpPtr + 1;
  tmpPtr++;
 }
 tmpPtr = name;
 while(*ptrToName != '.') *tmpPtr++ = *ptrToName++;
 *tmpPtr = 0;
 // 뤥
 snap = fopen(fullName, "rb");
 if(!snap)
 {
  printf( "Can't open file %s\n", fullName );
  return;
 }
 
 fseek(snap, 0L, SEEK_END);
 fileSize = ftell(snap);
 fseek(snap, 0L, SEEK_SET);
 
 if(!getSnapState())
 {
  puts("Invalid snapshot");
  fclose(snap);
  return;
 }
 
 writeRGSFile();
 // 뢠 ࠭
 strlwr(name);
 switch(type)
 {
  case SNA48:
            saveSNA48pages();
            break;
  case SNA128_1:
  case SNA128_2:
            saveSNA128pages();
            break;
  case Z80_1xx:
            saveZ80_1pages();
            break;
  default:
            saveZ80_23pages();
 }
 int i;
 for(i = 0; i < 8; i++)   // 塞 5 ࠭  ⠡ 㦠 ࠭
  if(pageTable[i] == 0) break;
 pageTable[i] = 0x15;
 // 襬 稪
 printf("\nLoader will be placed at address #%04x\n", loaderAddr);
 char ldrFileName[12];
 strupr(name);
 sprintf(ldrFileName, "%s.$B", name);
 FILE *ldrFile = fopen(ldrFileName, "wb");
 if(!ldrFile)
 {
  printf("Can't create loader %s\n", ldrFileName);
  return;
 }
 byte *loader;
 int offset = 0x74;
 if(ss.hmode == mode48)
 {
  loader = ldr48;
   if(pageTable[0] != 0x12)
   {
    loader[0x5a] = 0x18;
    loader[0x5b] = 0x06;
   }
  if(pageTable[0] != 0x10 && pageTable[1] != 0x10)
  {
   loader[0x62] = 0x18;
   loader[0x62] = 0x06;
  }
 }
 else
 {
  loader = ldr128;
  loader[0x6f] = ss.port7ffd;
  memcpy(loader+0xfa, pageTable, 8 );
 }
 
 // FIX'  ࠧ饭 稪  SP
 *(word*)(loader+0x96) = loaderAddr;
 *(word*)(loader+0x8f) = loaderAddr + loaderSize + 10; 

 // 㦠 ॣ

 *(word*)(loader+offset)    = ss.iy;
 *(word*)(loader+offset+4)  = ss._hl;
 *(word*)(loader+offset+7)  = ss._de;
 *(word*)(loader+offset+10) = ss._bc;
 loader[offset+14] = ss.bcolor;
 loader[offset+18] = ss._f;
 loader[offset+19] = ss._a;
 
 *(word*)(loader+offset+56)    = ss.ix;
 *(word*)(loader+offset+59) = ss.de;
 *(word*)(loader+offset+62) = ss.bc;
 loader[offset+65] = ss.i;
 ss.r &= 0x7F; ss.r -= 9; ss.r |= ss.bit7r;
 loader[offset+69] = ss.r;
 loader[offset+73] = ss.f;
 loader[offset+74] = ss.a;
 *(word*)(loader+offset+78) = ss.hl;
 *(word*)(loader+offset+81) = ss.sp;
 byte imode;
 switch(ss.imode)
 {
  case 0:
            imode  = 0x46;
            break;
  case 1:
            imode  = 0x56;
            break;
  case 2:
            imode  = 0x5E;
            break;
  default:
            puts( "Invalid interrupt mode" );
 }
 loader[offset+84] = imode;
 loader[offset+85] = ss.istatus ? 0xFB : 0xF3 ;
 if(type != SNA48)
  *(word*)(loader+offset+87) = ss.pc;
 else
 {
  loader[offset+86] = 0xED;
  loader[offset+87] = 0x45;
 }
 
 strncpy(loader, name, strlen(name));
 *(word*)(loader + 15) = checkSum(loader);
 // ⠥  128  48 稪  1 ᥪ
 fwrite(loader, sizeof(ldr48), 1, ldrFile);
 fclose(ldrFile);
 puts("Basic loader created");
}